#include "config.h"

#include <libgnome/libgnome.h>
#include <gtk/gtk.h>
#include <string.h>

#include <fontconfig/fontconfig.h>

#include "gdmwm.h"
#include "gdmcommon.h"
#include "gdmlanguages.h"
#include "greeter.h"
#include "greeter_configuration.h"
#include "greeter_item_pam.h"
#include "greeter_action_language.h"

#define LAST_LANGUAGE "Last"
#define DEFAULT_LANGUAGE "Default"

enum {
  LOCALE_COLUMN,
  TRANSLATED_NAME_COLUMN,
  UNTRANSLATED_NAME_COLUMN,
  IS_PREFERRED_COLUMN,
  NUM_COLUMNS
};

static GtkTreeModelFilter *filter_model = NULL;
static GtkListStore *lang_model = NULL;
static GtkWidget *dialog = NULL;
static gboolean savelang = FALSE;
static gchar *current_language = NULL;
static gchar *dialog_selected_language = NULL;
static gboolean show_all_languages = FALSE;

static char *
get_language_code_from_locale (const char *locale)
{
  char *language_code;
  char *end;

  language_code = g_strdup (locale);

  end = strchr (language_code, '_');

  if (end != NULL)
    {
      *end = '\0';
      return language_code;
    }

  end = strchr (language_code, '.');

  if (end != NULL)
    {
      *end = '\0';
      return language_code;
    }

  end = strchr (language_code, '@');

  if (end != NULL)
    {
      *end = '\0';
      return language_code;
    }

  return language_code;
}

static gboolean
locale_is_displayable (const char *locale)
{
  char *language_code;
  gboolean is_preferred;

  FcPattern *pattern;
  FcObjectSet *object_set;
  FcFontSet *font_set;

  is_preferred = FALSE;
  pattern = NULL;
  object_set = NULL;
  font_set = NULL;

  language_code = get_language_code_from_locale (locale);

  pattern = FcPatternBuild (NULL, FC_LANG, FcTypeString, language_code, NULL);

  if (pattern == NULL)
    goto done;

  object_set = FcObjectSetBuild (NULL, NULL);

  if (object_set == NULL)
    goto done;

  font_set = FcFontList (NULL, pattern, object_set);

  if (font_set == NULL)
    goto done;

  is_preferred = (font_set->nfont > 0);

done:

  if (font_set != NULL)
    FcFontSetDestroy (font_set);

  if (object_set != NULL)
    FcObjectSetDestroy (object_set);

  if (pattern != NULL)
    FcPatternDestroy (pattern);

  g_free (language_code);
  return is_preferred;
}

static gboolean
locale_is_preferred (const char *language_code)
{
  const GList *langauge_names, *tmp;
  const char * const * language_names;
  const char *preferred_language_list;
  gchar **preferred_language_codes;
  int i;
  gboolean is_preferred;

  language_names = ve_i18n_get_language_list ("LC_MESSAGES");

  tmp = language_names;
  while (tmp != NULL) {
	if (strcmp (language_code, tmp->data) == 0) {
	      return TRUE;
	}

        tmp = tmp->next;
  }

  preferred_language_list = g_getenv ("SUPPORTED");

  if (preferred_language_list == NULL)
    preferred_language_list = "en:de:es:fr:it:pt_BR:ja:ko:zh_CN:zh_TW:bn_IN:gu_IN:hi_IN:pa_IN:ta_IN";

  preferred_language_codes = g_strsplit (preferred_language_list, ":", 0);
  i = 0;
  is_preferred = FALSE;
  while (preferred_language_codes[i] != NULL) {
	if (g_str_has_prefix (language_code, preferred_language_codes[i])) {
	      is_preferred = TRUE;
	      break;
	}
	i++;
  }
  g_strfreev (preferred_language_codes);

  return is_preferred;
}

static gboolean
model_row_is_visible (GtkTreeModel *model, GtkTreeIter *iter)
{
  gchar *locale;
  gboolean is_visible = TRUE, is_preferred;

  gtk_tree_model_get (model, iter, 
                      LOCALE_COLUMN, &locale, 
                      IS_PREFERRED_COLUMN, &is_preferred, 
                      -1);

  if ((locale != NULL) && (strcmp (locale, "-") == 0))
    {
      is_visible = FALSE;
      goto out;
    }

  if (show_all_languages)
    {
      is_visible = TRUE;
      goto out;
    }

  is_visible = is_preferred;

out:
  g_free (locale);
  return is_visible;
}

static void
greeter_langauge_initialize_model (void)
{
  GList *list, *li;
  GtkTreeIter iter;
  GtkTreeIter separator_iter;

  list = gdm_lang_read_locale_file (GdmLocaleFile);

  lang_model = gtk_list_store_new (NUM_COLUMNS,
                                   G_TYPE_STRING,
                                   G_TYPE_STRING,
                                   G_TYPE_STRING,
                                   G_TYPE_BOOLEAN);

  filter_model = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (GTK_TREE_MODEL (lang_model), NULL));
  gtk_tree_model_filter_set_visible_func (filter_model,
                                          (GtkTreeModelFilterVisibleFunc) 
                                          model_row_is_visible, NULL, NULL);

  gtk_list_store_append (lang_model, &iter);
  gtk_list_store_set (lang_model, &iter,
		      TRANSLATED_NAME_COLUMN, _("Last"),
		      UNTRANSLATED_NAME_COLUMN, NULL,
		      LOCALE_COLUMN, LAST_LANGUAGE,
                      IS_PREFERRED_COLUMN, TRUE,
		      -1);

  gtk_list_store_append (lang_model, &iter);
  gtk_list_store_set (lang_model, &iter,
		      TRANSLATED_NAME_COLUMN, _("System Default"),
		      UNTRANSLATED_NAME_COLUMN, NULL,
		      LOCALE_COLUMN, DEFAULT_LANGUAGE,
                      IS_PREFERRED_COLUMN, TRUE,
		      -1);

  gtk_list_store_append (lang_model, &iter);
  gtk_list_store_set (lang_model, &iter,
                      TRANSLATED_NAME_COLUMN, NULL,
                      UNTRANSLATED_NAME_COLUMN, NULL,
                      LOCALE_COLUMN, "-",
                      -1);
  separator_iter = iter;

  for (li = list; li != NULL; li = li->next)
    {
      char *lang = li->data;
      char *name;
      char *untranslated;
      gboolean is_preferred;

      li->data = NULL;

      if (!locale_is_displayable (lang))
          continue;

      name = gdm_lang_name (lang,
			    FALSE /* never_encoding */,
			    TRUE /* no_group */,
			    FALSE /* untranslated */,
			    FALSE /* markup */);

      untranslated = gdm_lang_untranslated_name (lang,
						 TRUE /* markup */);

      if (untranslated == NULL)
        untranslated = g_strdup (name);
      else if (untranslated[0] == '\0')
        {
          g_free (untranslated);
          untranslated = g_strdup (name);
        }

      is_preferred = locale_is_preferred (lang);
      if (is_preferred)
        gtk_list_store_insert_before (lang_model, &iter, &separator_iter);
      else
        gtk_list_store_append (lang_model, &iter);
      gtk_list_store_set (lang_model, &iter,
			  TRANSLATED_NAME_COLUMN, name,
			  UNTRANSLATED_NAME_COLUMN, untranslated,
			  LOCALE_COLUMN, lang,
                          IS_PREFERRED_COLUMN, is_preferred,
			  -1);

      g_free (name);
      g_free (untranslated);
      g_free (lang);
    }
  g_list_free (list);
  gtk_tree_model_filter_refilter (filter_model);
}

gboolean
greeter_language_get_save_language (void)
{
  return savelang;
}

gchar *
greeter_language_get_language (const char *old_language)
{
  gchar *retval = NULL;

  /* Don't save language unless told otherwise */
  savelang = FALSE;

  if (old_language == NULL)
    old_language = "";

  /* If a different language is selected */
  if (current_language != NULL && strcmp (current_language, LAST_LANGUAGE) != 0)
    {
      if (strcmp (current_language, DEFAULT_LANGUAGE) == 0)
	retval = g_strdup ("");
      else
        retval = g_strdup (current_language);

      /* User's saved language is not the chosen one */
      if (strcmp (old_language, retval) != 0)
	{
	  gchar *msg;
	  char *current_name, *saved_name;

	  if (strcmp (current_language, DEFAULT_LANGUAGE) == 0)
	    current_name = g_strdup (_("System Default"));
	  else
	    current_name = gdm_lang_name (current_language,
					  FALSE /* never_encoding */,
					  TRUE /* no_group */,
					  TRUE /* untranslated */,
					  TRUE /* markup */);
	  if (strcmp (old_language, "") == 0)
	    saved_name = g_strdup (_("System Default"));
	  else
	    saved_name = gdm_lang_name (old_language,
					FALSE /* never_encoding */,
					TRUE /* no_group */,
					TRUE /* untranslated */,
					TRUE /* markup */);

	  msg = g_strdup_printf (_("You have chosen %s for this session, but your default setting is "
				   "%s.\nDo you wish to make %s the default for future sessions?"),
				 current_name, saved_name, current_name);
	  g_free (current_name);
	  g_free (saved_name);

	  savelang = gdm_common_query (msg, TRUE /* markup */, _("Make _Default"), _("Just For _This Session"));
	  g_free (msg);
	}
    }
  else
    {
      retval = g_strdup (old_language);
    }

  return retval;
}

static void
selection_changed (GtkTreeSelection *selection,
		   gpointer          data)
{
  GtkTreeIter iter;

  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
    {
      g_free (dialog_selected_language);
      gtk_tree_model_get (GTK_TREE_MODEL (filter_model), &iter, LOCALE_COLUMN, &dialog_selected_language, -1);
    }
}

static void
tree_row_activated (GtkTreeView         *view,
                    GtkTreePath         *path,
                    GtkTreeViewColumn   *column,
                    gpointer            data)
{
  GtkTreeIter iter;
  if (gtk_tree_model_get_iter (GTK_TREE_MODEL (lang_model), &iter, path))
    {
      g_free (dialog_selected_language);
      gtk_tree_model_get (GTK_TREE_MODEL (filter_model), &iter,
			  LOCALE_COLUMN, &dialog_selected_language,
			  -1);
      gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
  }
}

static void
more_langs_button_toggled (GtkToggleButton *expander)
{
  GtkWidget *image;
  gboolean is_active;

  if (GTK_BIN (expander)->child != NULL)
    gtk_widget_destroy (GTK_BIN (expander)->child);

  is_active = gtk_toggle_button_get_active (expander);

  image = gtk_image_new_from_stock (is_active? 
                                    GTK_STOCK_GO_DOWN : GTK_STOCK_GO_UP,
                                    GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image);
  gtk_container_add (GTK_CONTAINER (expander), image);

  show_all_languages = is_active;

  if (filter_model != NULL)
    gtk_tree_model_filter_refilter (filter_model);
}

static void 
lang_view_size_request (GtkWidget     *widget,
                        GtkRequisition *requisition)
{
  GtkWidget *dialog;
  int desired_width, desired_height;

  if (!GTK_WIDGET_MAPPED (widget))
    return;

  dialog = gtk_widget_get_toplevel (widget);

  desired_width = MIN (dialog->allocation.width - widget->allocation.width + requisition->width, 
                       (int) (.8 * gdm_wm_screen.width));
  desired_height = MIN (dialog->allocation.height - widget->allocation.height + requisition->height, 
                       (int) (.8 * gdm_wm_screen.height));

  if ((dialog->allocation.width == desired_width) && 
      (dialog->allocation.height == desired_height))
    return;

  gtk_window_resize (GTK_WINDOW (dialog), desired_width, desired_height);
}

void
greeter_action_language (GreeterItemInfo *info,
			 gpointer         user_data)
{
  GtkWidget *view = NULL;

  if (dialog == NULL)
    {
      GtkWidget **tmp_p;
      GtkWidget *swindow;
      GtkWidget *label;
      GtkWidget *expander;
      char *s;

      dialog = gtk_dialog_new_with_buttons (_("Select a language"),
#if TODO
					    GTK_WINDOW (parent_window),
#endif
					    NULL,
					    0,
					    GTK_STOCK_CANCEL,
					    GTK_RESPONSE_CANCEL,
					    GTK_STOCK_OK,
					    GTK_RESPONSE_OK,
					    NULL);
      gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER_ALWAYS);
      gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				       GTK_RESPONSE_OK);
      /* evil gcc warnings */
      tmp_p = &dialog;
      g_object_add_weak_pointer (G_OBJECT (dialog), (gpointer *)tmp_p);
      s = g_strdup_printf ("<span size=\"x-large\" weight=\"bold\">%s</span>",
			   _("Select a language for your session to use:"));
      label = gtk_label_new (s);
      g_free (s);
      gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
			  label, FALSE, FALSE, 6);
      view = gtk_tree_view_new ();

      g_signal_connect_after (G_OBJECT (view), "size-request",
                              G_CALLBACK (lang_view_size_request), NULL);
      gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (view), TRUE);
      /* FIXME: we should handle this better, but things really look
       * like crap if we aren't always LTR */
      gtk_widget_set_direction (view, GTK_TEXT_DIR_LTR);
      gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (view), FALSE);
      gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       GTK_DIALOG_MODAL,
					       NULL,
					       gtk_cell_renderer_text_new (),
					       "text", TRANSLATED_NAME_COLUMN,
					       NULL);
      gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
					       GTK_DIALOG_MODAL,
					       NULL,
					       gtk_cell_renderer_text_new (),
					       "markup",
					       UNTRANSLATED_NAME_COLUMN,
					       NULL);
      swindow = gtk_scrolled_window_new (NULL, NULL);
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
      gtk_container_add (GTK_CONTAINER (swindow), view);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
			  swindow, TRUE, TRUE, 0);
#if 0
      gtk_window_set_default_size (GTK_WINDOW (dialog),
				   MIN (400, gdm_wm_screen.width),
				   MIN (600, gdm_wm_screen.height));
#endif
      g_signal_connect (G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (view))),
			"changed",
			(GCallback) selection_changed,
			NULL);
      g_signal_connect(G_OBJECT (view),
                        "row_activated",
                        (GCallback) tree_row_activated,
                        NULL);

      expander = gtk_toggle_button_new ();

      g_signal_connect (G_OBJECT (expander),
                        "toggled", (GCallback) more_langs_button_toggled,
                        NULL);

      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (expander)) !=
          show_all_languages)
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (expander),
                                      show_all_languages);
      else
        more_langs_button_toggled (GTK_TOGGLE_BUTTON (expander));

      gtk_widget_show (expander);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
			  expander, FALSE, FALSE, 0);

      gtk_widget_show_all (dialog);
      gdm_wm_center_window (GTK_WINDOW (dialog));
    }
  gdm_wm_no_login_focus_push ();
  if (view != NULL)
    {
      gtk_widget_show_now (dialog);
      greeter_langauge_initialize_model ();
      gtk_tree_view_set_model (GTK_TREE_VIEW (view),
			       GTK_TREE_MODEL (filter_model));
    }
  switch (gtk_dialog_run (GTK_DIALOG (dialog)))
    {
    case GTK_RESPONSE_OK:
      if (dialog_selected_language)
	current_language = g_strdup (dialog_selected_language);
      break;
    case GTK_RESPONSE_CANCEL:
    default:
      break;
    }

  gdm_wm_no_login_focus_pop ();

  if (dialog)
    gtk_widget_hide (dialog);
}

